#ifndef __cuckoo_block_simple_h__
#define __cuckoo_block_simple_h__

#include "cuckoo-common.h"

template< class K, class HF >
class HashFunDouble : public HashFun<K> {
protected:
  HF m_q, m_h;
public:
  HashFunDouble( const unsigned long nBuckets = 1 ) :
  HashFun<int>( nBuckets ), m_q(nBuckets), m_h(nBuckets ){}

  /** Berechnet q(x)-i */
  inline typename HashFunDouble<K,HF>::HashVal
  q( const typename HashFunDouble<K,HF>::Key &s,
     const typename HashFunDouble<K,HF>::HashVal &i ) const{
    int res = m_q(s) - i;
    return ( res < 0 ) ? this->m_nBuckets+res : res;
  }

  /** Berechnet h1(x), h2(x) */
  inline pair<typename HashFunDouble<K,HF>::HashVal,
  typename HashFunDouble<K,HF>::HashVal> operator()(
    const typename HashFunDouble<K,HF>::Key &s ) const{
    int _h = m_h(s), _q = m_q(s);
    int r1 = _q - _h;
    if( r1 < 0 )
      r1 += this->m_nBuckets;
    return make_pair( _h, r1 );
  }
};



template< int deg = 2 >
class intStaticHashFun_poly_double :
  public HashFun<int> {
protected:
  const static INT64 m_bigPrime = 1073741827;
  INT64 m_h[deg+1];
  INT64 m_q[deg+1];
public:
  intStaticHashFun_poly_double(
    const unsigned long nBuckets = 1 ) :
  HashFun<int>( nBuckets ) {
    for( int j = deg; j >= 0; j-- ) {
      m_q[j] = my_rand( );
      m_h[j] = my_rand( );
    }
  }

  /** Berechnet q(x)-i */
  HashVal q( const Key &s, const HashVal &i ) const {
    INT64 ll = m_q[deg] * s + m_q[deg-1];
    for( int j = deg-2; j >= 0; j-- )
      ll = ( ll % m_bigPrime ) * s + m_q[j];
    int result = (int)( ll % m_bigPrime );
    return ( result - i ) % this->m_nBuckets;
  }

  /** Berechnet h1(x), h2(x) */
  pair<HashVal,HashVal> operator()( const Key &s ) const{
    INT64 ll0 = m_h[deg] * s + m_h[deg-1];
    INT64 ll1 = m_q[deg] * s + m_q[deg-1];
    for( int j = deg-2; j >= 0; j-- ) {
      ll0 = ( ll0 % m_bigPrime ) * s + m_h[j];
      ll1 = ( ll1 % m_bigPrime ) * s + m_q[j];
    }
    int r0 = (int)( ll0 % m_bigPrime );
    int r1 = (int)( ll1 % m_bigPrime );
    return make_pair(
             r0 % this->m_nBuckets,
             (r1 - r0) % this->m_nBuckets );
  }
};


template<class K, class HF, class EMPTY>
class CuckooBlockSimple :
  public CuckooDictionary<K, HF, EMPTY> {
protected:
  int d1;
  HF h;
  int nBlocks;

public:
  CuckooBlockSimple( int _d=4, const double &_epsilon=0.1 ):
  CuckooDictionary<K,HF,EMPTY>( _d, _epsilon ) {
    d1 = _d-1;
    reserve( 4*this->d );
  };

  int insert( const K &_x ) {
    K x( _x );
    int freecell = not_an_index;

    pair<int,int> p( h(x) );
    int r = binrand();
    int ind[2] = {p.first, p.second};
    int _i;
    int i0 = (_i = ind[r]) * this->d;
    int i1 = ind[1-r] * this->d;
    int i = i0;
    int ende = i0+this->d;

    for( ; i0 < ende; i0++, i1++ ) {
      if( this->mem[i0] == x )
        return found;
      else
        if( freecell == not_an_index &&
            this->__empty.is_empty(this->mem[i0]) )
          freecell = i0;
      if( this->mem[i1] == x )
        return found;
      else
        if( freecell == not_an_index &&
            this->__empty.is_empty(this->mem[i1]) )
          freecell = i1;
      if( this->__empty.is_empty(this->mem[i0]) &&
          this->__empty.is_empty(this->mem[i1]) )
        break;
    }

    this->account -= 2;

    if( freecell != not_an_index ) {
      this->mem[freecell] = x;
      this->m_nkeys++;
      return freecell;
    }

    while( this->account > 0 ) {
      int j = i + my_rand( ) % this->d;
      swap( x, this->mem[j] );
      r = binrand();
      _i = h.q( x, _i );
      i0 = _i * this->d;
      i = i0;
      int ende = i0+this->d;
      this->account--;
      for( ; i0 < ende; i0++ ) {
        if( this->__empty.is_empty(this->mem[i0]) ) {
          this->mem[i0] = x;
          this->m_nkeys++;
          return i0;
        }
      }
    }
    this->remaining_key = x;
    return not_an_index;
  }
  
  int insert_grow( const K & _x ) {
    this->account = this->insert_grow_account;
    int index = insert( _x );
    if( index != not_an_index )
      return index;
    if( !grow_by( this->f_grow ) )
      return not_an_index;
    return insert( this->remaining_key );
  };

  void clear( ) {
    for( int j = this->m-1; j >= 0; j-- )
      this->mem[ j ] = this->__empty;
    this->m_nkeys = 0;
    this->remaining_key = this->__empty;
  }

  /** Versucht, ncells Zellen zu reservieren 
  und die Schlssel dort unterzubringen. Wenn 
  das nicht geht, wird false zurckgegeben. */
  bool reserve( int ncells ) {
    HF _h = h;
    int oldm = this->m;
    int oldnkeys = this->m_nkeys;
    K old_remaining_key = this->remaining_key;
    int old_account = this->account;
    int old_nBlocks = nBlocks;

    nBlocks = (ncells+d1) / this->d;
    h = HF( nBlocks );
    ncells = nBlocks * this->d;

    K *oldmem = this->mem;
    this->mem = new K[ ncells ];
    this->m = ncells;
    this->set_account( (int)(this->account_factor * this->m) );
    this->clear( );
    this->remaining_key = old_remaining_key;
    if( oldmem != NULL ) {
      for( int j = oldm-1; j >= 0; j-- ) {
        if( !this->__empty.is_empty(oldmem[j]) ) {
          if( insert(oldmem[j]) == not_an_index ) {
            h = _h;
            delete[] this->mem;
            this->mem = oldmem;
            this->m = oldm;
            this->m_nkeys = oldnkeys;
            this->remaining_key = old_remaining_key;
            this->account = old_account;
            return false;
          }
        }
      }
      delete[] oldmem;
    }
    return true;
  };

  bool grow_by( const double &factor ) {
    int _m = (int)ceil( factor * this->m );
    if( factor > 1 && _m == this->m )
      _m++;
    return reserve( _m );
  }

  inline int lookup( const K &x ) const {
    pair<int,int> p( h(x) );
    p.first *= this->d;
    if( this->mem[p.first] == x )
      return p.first;
    p.first++;
    p.second *= this->d;
    int ende=p.first+this->d;
    if( this->mem[p.second] == x )
      return p.second;
    p.second++;
    for( ; p.first < ende; p.first++, p.second++ ) {
      if( this->mem[p.first] == x )
        return p.first;
      if( this->mem[p.second] == x )
        return p.second;
    }
    return not_an_index;
  }

  void remove( const K & x ) {
    pair<int,int> p( h(x) );
    p.first *= this->d;
    p.second *= this->d;
    int block_ende0 = p.first+this->d;
    int block_ende1 = p.second+this->d;
    int ende, zelle = not_an_index;
    for( ; p.first < block_ende0; p.first++, p.second++ ){
      if( this->mem[p.first] == x ) {
        ende = block_ende0;
        zelle = p.first;
        break;
      }
      if( this->mem[p.second] == x ) {
        ende = block_ende1;
        zelle = p.second;
        break;
      }
    }
    if( zelle == not_an_index )
      return;
    int _zelle = zelle;
    zelle++;
    while( zelle < ende &&
           !this->__empty.is_empty(this->mem[zelle]) )
      zelle++;
    zelle--; // Zelle zeigt auf den letzten belegten Index

    if( zelle > _zelle )
      this->mem[_zelle] = this->mem[zelle];
    this->mem[zelle] = this->__empty;
    this->m_nkeys--;
  }
};

#endif
