#ifndef __dictionary_common_h__
#define __dictionary_common_h__

using namespace std;

class int_empty_cell {
public:
  bool is_empty( const int &k ) const {
    return k == -1;
  };
  operator int() const {
    return -1;
  };
};


class pchar_empty_cell {
public:
  bool is_empty( const char* &k ) const {
    return k == (char *)0;
  };
  operator char*() const {
    return (char *)0;
  };
};


#define not_an_index (-1)
#define found (-2)

#include <stdlib.h>

#define my_rand lrand48

inline int binrand( ) {
  return my_rand( ) & 0x10 ? 1 : 0;
};

template< class K >
class HashFun {
protected:
  unsigned long m_nBuckets;
public:
  HashFun( const unsigned long nBuckets = 1 ) {
    m_nBuckets = nBuckets;
  };
  ~HashFun( ) {};
  typedef K Key;
  typedef long HashVal;
};

#ifdef __INTEL_COMPILER
typedef __int64 UINT64;
typedef __int64 INT64;
#else
typedef unsigned long long int UINT64;
typedef long long int INT64;
#endif


template< int deg = 2 >
class intStaticHashFun_poly : public HashFun<int> {
protected:
  const static UINT64 m_bigPrime = 1073741827;
  UINT64 m_a[deg+1];
public:
  intStaticHashFun_poly( const unsigned long nBuckets = 1 ):
  HashFun<int>( nBuckets ) {
    for( int j = deg; j >= 0; j-- )
      m_a[j] = my_rand( );
  }

  HashVal operator()( const Key &s ) const {
    UINT64 ll = m_a[deg] * s + m_a[deg-1];
    for( int j = deg-2; j >= 0; j-- )
      ll = ( ll % m_bigPrime ) * s + m_a[j];
    int result = (int)( ll % m_bigPrime );
    return result % m_nBuckets;
  }
};

template< int deg = 2 >
class intStaticHashFun_poly_offset :
  public intStaticHashFun_poly<deg> {
protected:
  int offset;
public:
  intStaticHashFun_poly_offset(
    const unsigned long nBuckets = 1,
    const unsigned int _offset = 0 ) :
  intStaticHashFun_poly<deg>( nBuckets ) {
    offset = _offset;
  }

  inline HashFun<int>::HashVal operator()(
    const HashFun<int>::Key &s ) const {
    int result = intStaticHashFun_poly<deg>::operator()(s);
    return result + offset;
  }
};


template <class K, class HF, class EMPTY>
class Dictionary {
protected:
  double epsilon;
  int m_nkeys;
  double f_grow;
  int m;
  K *mem;
  int account;

  /** Dieser Schlssel hat nicht gepasst */
  K remaining_key;

  EMPTY __empty;
public:
  Dictionary( const double &_epsilon=0.1 ) {
    remaining_key = __empty;
    set_epsilon( _epsilon );
    set_account(0);
    m_nkeys = 0;
    m = 0;
    mem = NULL;
    set_grow( 2 );
  };
  
  void clear( ) {
    for( int j = this->m-1; j >= 0; j-- )
      this->mem[ j ] = this->__empty;
    this->m_nkeys = 0;
    this->remaining_key = this->__empty;
  }

  ~Dictionary( ) {
    if( mem != NULL )
      delete mem;
  }

  inline void set_epsilon( const double & _epsilon ) {
    epsilon = _epsilon;
  };

  double get_loadfactor( ) const {
    return double(m) / m_nkeys;
  };

  void set_grow( const double &_f_grow ) {
    f_grow = _f_grow;
  }

  void set_account( const int _account ) {
    account = _account;
  };

  int get_account( ) const {
    return account;
  };

  int get_nkeys( ) const {
    return m_nkeys;
  };

  double get_avg_probes( ) const {
    return 1/epsilon + 1;
  };

  double get_element_probes( ) const {
    return 2/epsilon + 1;
  };
};

#endif
