#ifndef __lp_h__
#define __lp_h__

#include "dictionary-common.h"

template <class K, class HF, class EMPTY>
class LP : public Dictionary<K, HF, EMPTY> {
protected:
  HF h;
  int grow_at;

public:
  LP( const double &_epsilon=0.1 ) :
  Dictionary<K,HF,EMPTY>( _epsilon ) {
    reserve( 16 );
  };

  int insert( const K &_x ) {
    if( this->m_nkeys >= this->m ) {
      this->remaining_key = _x;
      return not_an_index;
    }
    int i = h(_x);
    /* Wir suchen das erste freie oder gelschte Feld
    oder brechen ab, falls _x gefunden wird */
    while( !this->__empty.is_empty(this->mem[i])
           && !(this->mem[i] == _x) ) {
      i--;
      if( i < 0 )
        i = this->m-1;
    }
    if( this->mem[i] == _x )
      return found;
    this->mem[i] = _x;
    this->m_nkeys++;
    return i;
  }

  void remove( const K & _x ) {
    int i = lookup( _x );
    if( i == not_an_index )
      return;
    this->mem[ i ] = this->__empty;
    int j = i;
    i--;
    if( i < 0 )
      i = this->m-1;
    while( !this->__empty.is_empty(this->mem[i]) ) {
      int r = h( this->mem[i] );
      if( ( r >= j || j >= i ) && ( j >= i || i > r )
          && ( i > r || r >= j ) ) {
        this->mem[j] = this->mem[i];
        this->mem[i] = this->__empty;
        j = i;
      }
      i--;
      if( i < 0 )
        i = this->m-1;
    }
    this->m_nkeys--;
  }

  int insert_grow( const K & _x ) {
    if( this->remaining_nkeys >= grow_at ) {
      grow_by( this->f_grow );
    }
    return insert( _x );
  };

  void clear( ) {
    for( int j = this->m-1; j >= 0; j-- )
      this->mem[ j ] = this->__empty;
    this->m_nkeys = 0;
    this->remaining_key = this->__empty;
  }

  /** Versucht, ncells Zellen zu reservieren
  und die Schlssel dort unterzubringen. 
  Wenn das nicht geht, wird false zurckgegeben. */
  bool reserve( int ncells ) {
    HF _h = h;
    int oldm = this->m;
    int oldnkeys = this->m_nkeys;
    K old_remaining_key = this->remaining_key;

    h = HF( ncells );
    K *oldmem = this->mem;
    this->mem = new K[ ncells ];
    this->m = ncells;
    this->clear( );
    this->remaining_key = old_remaining_key;
    if( oldmem != NULL ) {
      for( int j = oldm-1; j >= 0; j-- ) {
        if( oldmem[j] >= 0 ) {
          if( insert(oldmem[j]) == not_an_index ){
            h = _h;
            this->m = oldm;
            this->m_nkeys = oldnkeys;
            this->remaining_key = old_remaining_key;
            delete[] this->mem;
            this->mem = oldmem;
            return false;
          }
        }
      }
      delete[] oldmem;
    }
    return true;
  };

  bool grow_by( const double &factor ) {
    int _m = (int)ceil( factor * this->m );
    if( factor > 1 && _m == this->m )
      _m++;
    return reserve( _m );
  }

  inline int lookup( const K &_x ) const {
    int i = h(_x);
    while( !this->__empty.is_empty(this->mem[i]) ){
      if( this->mem[i] == _x )
        return i;
      i--;
      if( i < 0 )
        i = this->m-1;
    }
    return not_an_index;
  }
};

#endif
