#include "cuckoo-lp.h"
#include "cuckoo-dary.h"
#include "cuckoo-block.h"
#include "cuckoo-block-simple.h"
#include "lp.h"

#include "keygen.h"

#include <sys/times.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <iostream>
#include <algorithm>
#include <math.h>

using namespace std;

int nkeys = 10000;
double epsilon = 1;
int verfahren = 0;
int d = 8;

vector<int> keys;

tms tms_start, tms_stop;

inline void start( ) {
  times( &tms_start );
};

inline clock_t stop( ) {
  times( &tms_stop );
  return tms_stop.tms_stime + tms_stop.tms_utime
         - tms_start.tms_stime - tms_start.tms_utime;
};

template < class DICT >
void test_insert( DICT dict ) {
  clock_t t;

  double account_faktor = dict.get_avg_probes( );
  int the_account = (int)(account_faktor*nkeys);
  dict.reserve( (int) ((1+epsilon) * nkeys) );
  dict.set_account( the_account );
  clog << "Einfgen bei Platz " << 1+epsilon
  << " [account = " << the_account << "]" << endl;
  start();
  int index = 0;
  for( int i = keys.size( )-1; i >= 0; i-- ) {
    index = dict.insert( keys[i] );
    if( index == not_an_index )
      break;
  }
  t = stop( );
  if( index == not_an_index )
    cout << "Zeit (alle) [sek]: ***" << endl;
  else
    cout << "Zeit (alle) [sek]: "
    << double(t) / sysconf( _SC_CLK_TCK ) << endl;
  clog << "Restkonto: " << dict.get_account( ) << endl;

  clog << "Loadfactor: " << dict.get_loadfactor( ) << endl;
  clog << "Positive Lookups: ";
  start();
  for( int i = keys.size( )-1; i >= 0; i-- ) {
    if( dict.lookup( keys[i] ) == -1 ) {
      cerr << "Fehler beim Nachschlagen!" << endl;
      clog << flush;
      exit( -1 );
    }
  }
  t = stop( );
  cout << "Pos. Lookups [sek]: " <<
  double(t) / sysconf( _SC_CLK_TCK ) << endl;

  clog << "Negative Lookups: ";
  /* Wir testen hier nkeys Schlssel,
  die garantiert nicht im Wrterbuch sind.*/
  int neg = *(keys.end( )-1)+1;
  start();
  for( int i = keys.size( )-1; i >= 0; i--, neg++ ) {
    if( dict.lookup( neg ) != -1 ) {
      cerr << "Fehler beim Nachschlagen!" << endl;
      clog << flush;
      exit( -1 );
    }
  }
  t = stop( );
  cout << "Neg. Lookups [sek]: "
  << double(t) / sysconf( _SC_CLK_TCK ) << endl;

  account_faktor = dict.get_element_probes( );
  the_account = (int)(account_faktor*log((double)nkeys));
  clog << "Lschen/Einfgen: "
  << "[account = " << the_account << "]" << endl;
  neg = *(keys.end( )-1)+1;
  start();
  for( int i = keys.size( )-1; i >= 0; i--, neg++ ) {
    dict.remove( keys[i] );
    dict.set_account( the_account );
    index = dict.insert( neg );
    if( index == not_an_index )
      break;
  }
  t = stop( );
  if( index == not_an_index )
    cout << "Zeit (remove/insert) [sek]: ***" << endl;
  else
    cout << "Zeit (remove/insert) [sek]: "
    << double(t) / sysconf( _SC_CLK_TCK ) << endl;
  clog << "Gre des Wrterbuches: "
  << dict.get_nkeys( ) << endl;

  clog << "Lookups nach remove/insert: ";
  neg = *(keys.end( )-1)+1;
  start();
  for( int i = keys.size( )-1; i >= 0; i--, neg++ ) {
    if( ( dict.lookup( neg ) == not_an_index ) ||
        ( dict.lookup( keys[i] ) != not_an_index ) ) {
      cerr << "Fehler beim Nachschlagen!" << endl;
      clog << flush;
      exit( -1 );
    }
  }
  t = stop( );
  cout << "Lookups nach remove [sek]: "
  << double(t) / sysconf( _SC_CLK_TCK ) << endl;
}


const int HF_DEG = 3;

#define CUCKOO_LP 1
#define CUCKOO_DARY 2
#define CUCKOO_BLOCK 3
#define CUCKOO_BLOCK_SIMPLE 4
#define DICT_LP 5


typedef intStaticHashFun_poly_offset<HF_DEG> HF_offset;
typedef intStaticHashFun_poly<HF_DEG> HF;

int main(int argc, char *argv[]) {
  if( argc < 2 ) {
    exit( -1 );
  }
  time_t t;
  srand48( time( &t ) );
  if( argc >= 2 )
    verfahren = atoi( argv[ 1 ] );
  if( argc >= 3 )
    nkeys = atoi( argv[ 2 ] );
  if( argc >= 4 )
    epsilon = atof( argv[ 3 ] );
  if( argc >= 5 )
    d = atoi( argv[ 4 ] );

  keys = KeyGen<int>(nkeys)();
  switch( verfahren ) {

  case CUCKOO_DARY :
    test_insert( CuckooDary<int, HF_offset,
                 int_empty_cell> ( d, epsilon ) );
    break;
  case CUCKOO_BLOCK :
    test_insert( CuckooBlock<int, HF,
                 int_empty_cell> ( d, epsilon ) );
    break;
  case CUCKOO_BLOCK_SIMPLE :
    test_insert( CuckooBlockSimple<int,HashFunDouble<int,HF>,
                 int_empty_cell> ( d, epsilon ) );
    break;
  case CUCKOO_LP :
    test_insert( CuckooLP<int, HF,
                 int_empty_cell>( d, epsilon ) );
    break;
  case DICT_LP :
    test_insert( LP<int, HF,
                 int_empty_cell> ( epsilon ) );
    break;
  }
  return 0;
}
