#ifndef __KEYGEN_H
#define __KEYGEN_H

#include <vector>
#include <iostream>

/** Ein Generator fr eindeutige Schlssel */
template< class _KT >
class KeyGen {
public:
  typedef _KT KeyType;

protected:
  int n; //! Anzahl der zu erzeugenden Schlssel

  /** Durchsucht das zunchst sortierte Feld
      nach doppelten Eintrgen. Wenn ein doppelter
      Eintrag gefunden wird, dann wir er durch 
      den nchsten verschiedenen Eintrag berschrieben.
      Gibt einen Iterator auf das Ende des Feldes
      mit paarweise verschiedenen Eintrgen zurck.
  */
  typename vector<KeyType>::iterator
  remove_doubles( typename vector<KeyType>::iterator first,
                  typename vector<KeyType>::iterator last) {
    if( first == last )
      return last;
    stable_sort( first, last );
    typename vector<KeyType>::iterator eq( first );
    typename vector<KeyType>::iterator dest( first+1 );
    for( ; first != last; first++ ) {
      if( !( *first == *eq ) ) {
        if( dest != first )
          *dest = *first;
        dest++;
        eq = first;
      }
    }
    return dest;
  };

public:
  /** Erzeugt einen Generator fr _n Schlssel */
  KeyGen( int _n ) {
    n = _n;
  };

  /** Erzeugt n paarwerise verschiedene Schlssel */
  vector<KeyType> operator( )( ) {
    vector<KeyType> keys( 13 * n / 10 );
    typename vector<KeyType>::iterator it;
    for( it = keys.begin( ); it != keys.end( ); it++ )
      *it = lrand48( ) >> 4;
    it = remove_doubles( keys.begin( ), keys.end( ) );
    int k = it - keys.begin( );
    /* Test, ob gengend Schlssel erzeugt worden sind */
    if( k < n ) {
      cerr << "Nicht gengend Schlssel erzeugt!" << endl;
      exit( -1 );
    }
    keys.resize( n );
    return keys;
  }
};

#endif
